import asyncio
import logging
import sys

import aiohttp
import aiopg
import marshmallow
import psycopg2

from funkwhale_network import exceptions, schemas, settings
from funkwhale_network.db import DB

logger = logging.getLogger(__name__)


def get_session_kwargs():
    headers = {"User-Agent": settings.CRAWLER_USER_AGENT}
    return {
        "timeout": aiohttp.ClientTimeout(total=settings.CRAWLER_TIMEOUT),
        "headers": headers,
    }


async def fetch_nodeinfo(session, domain):
    nodeinfo = await get_well_known_data(session, domain)
    data = await get_nodeinfo(session, nodeinfo)
    return data


async def get_well_known_data(session, domain, protocol="https"):
    url = f"https://{domain}/.well-known/nodeinfo"  # noqa
    response = await session.get(url, ssl=False)
    return await response.json()


async def get_nodeinfo(session, nodeinfo):
    for link in nodeinfo.get("links", []):
        if link["rel"] == "http://nodeinfo.diaspora.software/ns/schema/2.0":
            response = await session.get(link["href"], ssl=False)
            return await response.json()
    raise exceptions.NoNodeInfo()


async def check(session, domain, stdout=sys.stdout):
    async with DB() as db:
        await db.create_domain({"name": domain})
        check_data = {"up": True, "domain": domain}
        try:
            nodeinfo = await fetch_nodeinfo(session, domain)
            cleaned_nodeinfo = await clean_nodeinfo(nodeinfo, domain_name=domain)
            cleaned_check = clean_check(check_data, cleaned_nodeinfo)
        except (aiohttp.client_exceptions.ClientError, exceptions.CrawlerError) as e:
            stdout.write(
                f"Error while fetching {domain}: {e}, marking instance as down"
            )
            check_data["up"] = False
            cleaned_check = check_data

        await db.save_check(cleaned_check)


async def crawl_all(session, *domains, stdout, max_passes):
    data = {
        "pending_domains": set(domains),
        "valid_domains": set(),
        "invalid_domains": set(),
        "handled_domains": set(),
        "results": {},
        "pass_number": 0,
    }

    def print_pass():
        stdout(
            f"[Pass {data['pass_number']}] "
            f"{len(data['pending_domains'])} new domains to crawl, "
            f"{len(data['handled_domains'])} checked, "
            f"{len(data['valid_domains'])} valid"
        )

    while data["pending_domains"] and data["pass_number"] < max_passes:
        data["pass_number"] += 1
        print_pass()
        domain_list = list(data["pending_domains"])
        tasks = [
            asyncio.create_task(crawl_single(session, domain, data, stdout))
            for domain in domain_list[:100]  # limit to 100 domains per pass
        ]
        await asyncio.gather(*tasks)  # Wait for all tasks to complete

    if data["pass_number"] < max_passes:
        print_pass()

    return data


async def crawl_single(session, domain, data, stdout):
    stdout(f"Starting crawl with {domain} seed domains…")
    try:
        nodeinfo_data = await fetch_nodeinfo(session, domain)
        cleaned_data = clean_check(
            {"domain": domain, "up": True}, await clean_nodeinfo(nodeinfo_data, domain)
        )
    except Exception as exception:
        data["invalid_domains"].add(domain)
        logger.debug(exception)
        stdout(f"Error while fetching {domain}: {exception}")
        return
    finally:
        data["pending_domains"].remove(domain)
        data["handled_domains"].add(domain)

    nodes_endpoint = recursive_getattr(nodeinfo_data, "metadata.endpoints.knownNodes")
    if nodes_endpoint:
        try:
            await gather_known_nodes(
                session, "https://" + domain + nodes_endpoint, data, stdout
            )
        except Exception as exception:
            logger.debug(exception)

    data["valid_domains"].add(domain)
    stdout(f"Checked valid domain {domain} successfully")
    data["results"][domain] = cleaned_data


async def gather_known_nodes(session, url, data, stdout):
    fetch_url = url
    while fetch_url:
        response = await session.get(fetch_url, ssl=False)
        result = await response.json()
        fetch_url = result.get("next")
        known_domains = {d["name"] for d in result["results"]}
        data["pending_domains"] |= known_domains - data["handled_domains"]


async def clean_nodeinfo(data, domain_name):
    schema = schemas.NodeInfo2Schema()
    try:
        result = schema.load(data)
    except marshmallow.exceptions.ValidationError as e:
        if (
            isinstance(e.messages, dict)
            and isinstance(e.messages.get("software"), dict)
            and (
                "Must be one of: funkwhale, Funkwhale"
                in e.messages.get("software", {}).get("name", [""])[0]
            )
        ):
            # to do : when we get domain from nodeinfo they are not on the db. Maybe we should create a periodic
            # task that delete wrong domain in the domain table instead of this
            async with DB() as db:
                await db.delete_domain(domain_name)
                print(
                    f"Deleted {domain_name} from database since it's not a Funkwhale instance",
                    flush=True,
                )
                return
        else:
            raise e
    return result


def recursive_getattr(obj, key, permissive=True):
    """
    Given a dictionary such as {'user': {'name': 'Bob'}} and
    a dotted string such as user.name, returns 'Bob'.

    If the value is not present, returns None
    """
    value = obj
    for k in key.split("."):
        try:
            value = value.get(k)
        except (TypeError, AttributeError):
            if not permissive:
                raise
            return
        if value is None:
            return

    return value


def clean_check(check_data, nodeinfo_data):
    return {
        "domain": check_data["domain"],
        "up": check_data["up"],
        "node_name": recursive_getattr(nodeinfo_data, "metadata.nodeName"),
        "open_registrations": recursive_getattr(nodeinfo_data, "openRegistrations"),
        "federation_enabled": recursive_getattr(
            nodeinfo_data, "metadata.library.federationEnabled"
        ),
        "anonymous_can_listen": recursive_getattr(
            nodeinfo_data, "metadata.library.anonymousCanListen"
        ),
        "private": recursive_getattr(nodeinfo_data, "metadata.private"),
        "usage_users_total": recursive_getattr(nodeinfo_data, "usage.users.total"),
        "usage_users_active_half_year": recursive_getattr(
            nodeinfo_data, "usage.users.activeHalfyear"
        ),
        "usage_users_active_month": recursive_getattr(
            nodeinfo_data, "usage.users.activeMonth"
        ),
        "usage_listenings_total": recursive_getattr(
            nodeinfo_data, "metadata.usage.listenings.total"
        ),
        "usage_downloads_total": recursive_getattr(
            nodeinfo_data, "metadata.usage.downloads.total"
        ),
        "library_tracks_total": recursive_getattr(
            nodeinfo_data, "metadata.library.tracks.total"
        ),
        "library_albums_total": recursive_getattr(
            nodeinfo_data, "metadata.library.albums.total"
        ),
        "library_artists_total": recursive_getattr(
            nodeinfo_data, "metadata.library.artists.total"
        ),
        "library_music_hours": recursive_getattr(
            nodeinfo_data, "metadata.library.music.hours"
        ),
        "software_name": recursive_getattr(nodeinfo_data, "software.name"),
        "software_version_major": recursive_getattr(
            nodeinfo_data, "software.version.major"
        ),
        "software_version_minor": recursive_getattr(
            nodeinfo_data, "software.version.minor"
        ),
        "software_version_patch": recursive_getattr(
            nodeinfo_data, "software.version.patch"
        ),
        "software_prerelease": recursive_getattr(
            nodeinfo_data, "software.version.prerelease"
        ),
        "software_build": recursive_getattr(nodeinfo_data, "software.version.build"),
    }


async def save_check(data):
    conn = await aiopg.connect(settings.DB_DSN)
    cursor = await conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
    node_name = data.pop("node_name", None)
    fields, values = [], []
    for field, value in data.items():
        fields.append(field)
        values.append(value)

    sql = "INSERT INTO checks (time, {}) VALUES (NOW(), {}) RETURNING *".format(
        ", ".join(fields), ", ".join(["%s" for _ in values])
    )
    await cursor.execute(sql, values)
    obj = await cursor.fetchone()

    if data.get("private") is True:
        # let's clean previous checks
        sql = "DELETE FROM checks WHERE domain = %s"
        await cursor.execute(sql, [data["domain"]])
        return
    if node_name:
        await cursor.execute(
            "UPDATE domains SET node_name = %s WHERE name = %s",
            [node_name, data["domain"]],
        )

    return obj
